/****************************************************************************
 * bot.cpp
 *  Copyright © 2011, Vsevolod Velichko.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/


#include "bot.h"
#include <jreen/mucroom.h>
#include <jreen/jid.h>
#include <jreen/disco.h>
#include <jreen/message.h>
#include <jreen/util.h>
#include <jreen/nickname.h>
#include <jreen/iq.h>
#include <jreen/ping.h>
#include <jreen/softwareversion.h>
#include <jreen/vcard.h>
#include <jreen/entitytime.h>
#include <QDebug>
#include <QDateTime>
#include <QTimeZone>
#include <QXmlQuery>
#include <QSqlError>
#include "handlers.h"
#include "settings.h"
#include "delayedmessage.h"
#ifdef WITH_MORPHO
#include "ortho.h"
#endif


static void cleanupDelayedMessages(QHash<QString, DelayedMessages> &messages)
{
	for (auto &it : messages.keys())
		if (!messages[it]->size())
			messages.remove(it);
}

static inline QString statusOfPresence(const Jreen::Presence &presence)
{
	QString status = !presence.status().isEmpty() ? QString::fromUtf8(" (%1)").arg(presence.status()) : "";
	switch(presence.subtype())
	{
		case Jreen::Presence::Type::Unavailable:
			return QString::fromUtf8("вышел из конференции") + status;
		case Jreen::Presence::Type::Available:
			return QString::fromUtf8("в сети") + status;
		case Jreen::Presence::Type::Chat:
			return QString::fromUtf8("готов к чату") + status;
		case Jreen::Presence::Type::Away:
			return QString::fromUtf8("отошёл") + status;
		case Jreen::Presence::Type::DND:
			return QString::fromUtf8("недоступен") + status;
		case Jreen::Presence::Type::XA:
			return QString::fromUtf8("совсем отошёл") + status;
		default:
			return QString();
	}
}

static inline QString stringOfAffiliation(const Jreen::MUCRoom::Affiliation affiliation)
{
	switch(affiliation)
	{
		case Jreen::MUCRoom::AffiliationMember:
			return QString::fromUtf8("пользователь");
		case Jreen::MUCRoom::AffiliationOwner:
			return QString::fromUtf8("владелец");
		case Jreen::MUCRoom::AffiliationAdmin:
			return QString::fromUtf8("администратор");
		default:
			return QString::fromUtf8("мутный тип");
	}
}

static inline QString stringOfRole(const Jreen::MUCRoom::Role role)
{
	switch(role)
	{
		case Jreen::MUCRoom::RoleVisitor:
			return QString::fromUtf8("посетитель");
		case Jreen::MUCRoom::RoleParticipant:
			return QString::fromUtf8("член");
		case Jreen::MUCRoom::RoleModerator:
			return QString::fromUtf8("модератор");
		default:
			return QString::fromUtf8("хуй");
	}
}

static void insertLogRecord(const QString &type, const QString &source, const QString &message, const QString &role = QString())
{
	QSqlQuery query;
	QDateTime ts = QDateTime::currentDateTime();
	query.prepare("INSERT INTO log (year, month, day, time, type, source, role, message) VALUES (:year, :month, :day, :time, :type, :source, :role, :message)");
	query.bindValue(":year", ts.date().year());
	query.bindValue(":month", ts.date().month());
	query.bindValue(":day", ts.date().day());
	query.bindValue(":time", ts.time().toString("hh:mm:ss.zzz"));
	query.bindValue(":type", type);
	query.bindValue(":source", source);
	query.bindValue(":role", role);
	query.bindValue(":message", message);
	query.exec();
	auto error = query.lastError();
	if (error.isValid())
	{
		qDebug() << "Error type: " << error.type();
		qDebug() << "Database: " << error.databaseText();
		qDebug() << "Driver: " << error.driverText();
	}
}

Bot::Bot()
	: client(new Jreen::Client(Jreen::JID(Options::instance()["botjid"].toString()), Options::instance()["botpassword"].toString()))
	, handlers(new Handlers(this))
	, db(QSqlDatabase::addDatabase("QSQLITE"))
{
	db.setDatabaseName(Options::instance()["database"].toString());
	db.open();
	db.exec("CREATE TABLE IF NOT EXISTS log (year INTEGER(2) NOT NULL, month INTEGER(1) NOT NULL, day INTEGER(1) NOT NULL, time VARCHAR(16) NOT NULL, type VARCHAR(16) NOT NULL, source TEXT NOT NULL, role TEXT, message TEXT NOT NULL)");
	db.exec("CREATE INDEX IF NOT EXISTS ymd ON log (year, month, day)");
	db.exec("CREATE INDEX IF NOT EXISTS ym ON log (year, month)");
	db.exec("CREATE INDEX IF NOT EXISTS y ON log (year)");

	delayedMessages = Settings::instance().get<QHash<QString,DelayedMessages> >("delayed_messages");
	cleanupDelayedMessages(delayedMessages);
	currentNicks = Settings::instance().get<QHash<QString,QString> >("current_nicks");
	nicksHistory = Settings::instance().get<QHash<QString,QString> >("nicks_history");
	client->setParent(this);
	connect(client, SIGNAL(serverFeaturesReceived(QSet<QString>)), this, SLOT(onConnected()));
	connect(client, SIGNAL(serverFeaturesReceived(QSet<QString>)), handlers, SLOT(onConnected()));

	connect(client, SIGNAL(disconnected(Jreen::Client::DisconnectReason)), this, SLOT(onDisconnected(Jreen::Client::DisconnectReason)));
	connect(this, SIGNAL(ready()), this, SLOT(joinGroupchat()));
	client->connectToServer();
}

void Bot::onConnected()
{
	insertLogRecord("connection", "system", "соединение с сервером установлено");
	roster = new Jreen::AbstractRoster( client );
	client->setPresence( Jreen::Presence::Available, "Pray or die", 15 );
	client->disco()->setSoftwareVersion( "Monastery", "1.0", "Holy Debian" );
	client->disco()->identities().append( Jreen::Disco::Identity( "client", "type", "Aaron", "en" ) );
	client->disco()->identities().append( Jreen::Disco::Identity( "client", "type", QString::fromUtf8("Аарон"), "ru" ) );
	client->setPingInterval(60000);
	roster->load();
	emit ready();
}

void Bot::onDisconnected(Jreen::Client::DisconnectReason /*reason */)
{
	insertLogRecord("connection", "system", "соединение с сервером разорвано");
	roster->deleteLater();
	client->connectToServer();
}

void Bot::joinGroupchat()
{
	room = new Jreen::MUCRoom(client, Jreen::JID(fullConferenceName()));
	room->setHistorySeconds(0);
	room->join();
	insertLogRecord("connection", "system", QString::fromUtf8("вход в конференцию: %1").arg(fullConferenceName()));
	connect(room, SIGNAL(messageReceived(Jreen::Message,bool)), this, SLOT(processMessage(Jreen::Message)));
	connect(room, SIGNAL(presenceReceived(const Jreen::Presence&,const Jreen::MUCRoom::Participant*)), this, SLOT(processPresence(const Jreen::Presence&,const Jreen::MUCRoom::Participant*)));
	connect(room, SIGNAL(subjectChanged(const QString&,const QString&)), this, SLOT(processSubject(const QString&,const QString&)));
	//sendPublicMessage("Мир вам, дети мои");
}

bool Bot::isPresent(const QString &nick)
{
	return room->realJid(nick).isValid() // directly to nick
		|| room->realJid(currentNicks[nick]).isValid() // 'nick' is JID really
		|| room->realJid(currentNicks[nicksHistory[nick]]).isValid(); // user has other nick at the moment
}

Bot::OwnerStatus Bot::getOwnerStatus(const QString &nick)
{
	static QStringList superOwners = Options::instance()["superOwners"].toStringList();
	static QStringList owners = Options::instance()["owners"].toStringList();
	qDebug() << "Getting jid of " << nick;
	QString user = room->realJid(nick).bare();
	qDebug() << "Jid is " << user;
	if (superOwners.contains(user))
		return SuperOwner;
	if (owners.contains(user))
		return Owner;
	return NonOwner;
}

bool Bot::isBot(const QString& nick)
{
	static QStringList bots = Options::instance()["bots"].toStringList();
	QString user = room->realJid(nick).bare();
	return bots.contains(user);
}

void Bot::ban (const QString &nick)
{
	qDebug() << "Banning the" << nick;
	insertLogRecord("action", "system", QString::fromUtf8("бан: %1").arg(nick));
	room->ban(nick, "Удостоен аудиенции у Бога.");
}

void Bot::unban (const QString &jid)
{
	qDebug() << "Unbanning the" << jid;
	insertLogRecord("action", "system", QString::fromUtf8("разбан: %1").arg(jid));
	room->setAffiliation(Jreen::JID(jid), Jreen::MUCRoom::AffiliationNone, "милосердие превыше всего.");
}

void Bot::kick (const QString &nick, const QString &reason)
{
	insertLogRecord("action", "system", QString::fromUtf8("кик: %1").arg(nick));
	room->kick(nick, reason);
}

void Bot::setModer (const QString &nick)
{
	qDebug() << "Making " << nick << "the moderator";
	room->setRole(nick, Jreen::MUCRoom::RoleModerator, "так надо");
}

void Bot::sendPing(const QString &nick)
{
	Jreen::JID jid;
	if (room->realJid(nick).isValid())
		jid.setJID(conferenceName() + "/" + nick);
	else
		jid.setJID(nick);
	Jreen::IQ iq(Jreen::IQ::Get, jid);
	iq.addExtension(new Jreen::Ping);
	client->send(iq, this, SLOT(onPingResult(const Jreen::IQ&,int)), 0);
	pings[iq.id()] = new PingRequest(nick);
}
void Bot::sendVersionRequest(const QString &nick)
{
	Jreen::JID jid;
	if (room->realJid(nick).isValid())
		jid.setJID(conferenceName() + "/" + nick);
	else
		jid.setJID(nick);
	Jreen::IQ iq(Jreen::IQ::Get, jid);
	iq.addExtension(new Jreen::SoftwareVersion);
	client->send(iq, this, SLOT(onVersionResult(const Jreen::IQ&,int)), 0);
	simpleRequests[iq.id()] = nick;
}
void Bot::sendVcardRequest(const QString &nick)
{
	Jreen::IQ iq(Jreen::IQ::Get, Jreen::JID(conferenceName() + "/" + nick));
	iq.addExtension(new Jreen::VCard);
	client->send(iq, this, SLOT(onVcardResult(const Jreen::IQ&,int)), 0);
	simpleRequests[iq.id()] = nick;
}

void Bot::sendTimeRequest(const QString &nick)
{
	Jreen::IQ iq(Jreen::IQ::Get, Jreen::JID(conferenceName() + "/" + nick));
	iq.addExtension(new Jreen::EntityTime);
	client->send(iq, this, SLOT(onTimeResult(const Jreen::IQ&,int)), 0);
	simpleRequests[iq.id()] = nick;
}

void Bot::sendInfoRequest(const QString &nick, const Jreen::JID &from)
{
	Jreen::JID jid;
	if (room->realJid(nick).isValid())
		jid.setJID(conferenceName() + "/" + nick);
	else
		jid.setJID(nick);
	Jreen::Disco::Item item(jid, QString(), QString());
	Jreen::DiscoReply *reply = client->disco()->requestInfo(item);
	connect(reply, SIGNAL(finished()), SLOT(onInfoResult()));
	reply->setProperty("nick", nick);
	reply->setProperty("from", from.full());
}

void Bot::onPingResult(const Jreen::IQ &response, int)
{
	PingRequest *r = pings.take(response.id());
	bool isError = false;
	if (response.error())
	{
		Jreen::Error::Condition c = response.error()->condition();
		switch(c)
		{
			case Jreen::Error::FeatureNotImplemented:
			case Jreen::Error::NotAllowed:
			case Jreen::Error::NotAcceptable:
			case Jreen::Error::NotAuthorized:
			case Jreen::Error::PaymentRequired:
			case Jreen::Error::RegistrationRequired:
			//case Jreen::Error::SubscribtionRequired:
			case Jreen::Error::Undefined:
				isError = true;
				break;
			default:
				sendPublicMessage(QString("%1 — неведом людям").arg(r->nick));
				delete r;
				return;
		}
	}
	handlers->parsePingResult(
			r->nick,
			r->timer.elapsed(),
			isError);
	delete r;
}
void Bot::onVersionResult(const Jreen::IQ &response, int)
{
	const QString &nick = simpleRequests.take(response.id());
	if (response.subtype() == Jreen::IQ::Error)
	{
		sendPublicMessage(QString("%1 — чзх?").arg(nick));
		return;
	}
	Jreen::SoftwareVersion::Ptr version = response.payload<Jreen::SoftwareVersion>();
	if (!version)
		handlers->parseVersionResult(nick, QString(), QString(), QString());
	else
		handlers->parseVersionResult(nick, version->name(), version->version(), version->os());
}
void Bot::onVcardResult(const Jreen::IQ &response, int)
{
	const QString &nick = simpleRequests.take(response.id());
	if (response.subtype() == Jreen::IQ::Error)
	{
		sendPublicMessage(QString("%1 — и от этого вы ждёте вкард?").arg(nick));
		return;
	}
	Jreen::VCard::Ptr vcard = response.payload<Jreen::VCard>();
	handlers->parseVcardResult(nick, vcard);
}

void Bot::onTimeResult(const Jreen::IQ &response, int)
{
	const QString &nick = simpleRequests.take(response.id());
	if (response.subtype() == Jreen::IQ::Error || !response.containsPayload<Jreen::EntityTime>())
	{
		sendPublicMessage(QString("и ходил %1 пред Богом; и не стало его, потому что Бог взял его.").arg(nick));
		return;
	}
	Jreen::EntityTime::Ptr time = response.payload<Jreen::EntityTime>();

	sendPublicMessage(QString("на часах %1 сейчас %2, в часовом поясе %3")
					  .arg(nick)
					  .arg(time->localTime().toString())
					  .arg(time->localTime().timeZoneAbbreviation())
					  );
}

void Bot::onInfoResult()
{
	Jreen::DiscoReply *reply = qobject_cast<Jreen::DiscoReply*>(sender());
	QString nick = reply->property("nick").toString();
	QString from = reply->property("from").toString();

	if (!xepsTimer.isValid() || xepsTimer.hasExpired(24 * 60 * 60 * 1000))
	{
		QUrl request(QString("http://xmpp.org/registrar/disco-features.xml"));
		QXmlQuery query;
		QStringList result;
		query.setFocus(request);
		query.setQuery(
		            "for $x in /registry/var/name | /registry/var/desc | /registry/var/doc\n"
		            "return string($x)");
		Q_ASSERT(query.isValid());
		query.evaluateTo(&result);
		for (int i = 0; i + 1 < result.size(); i += 3)
		{
			QString feature = result[i];
			QString desc = result[i + 1];
			QString doc = result[i + 2];
			xeps[feature] = qMakePair(desc, doc);
		}
		if (result.size() > 2)
			xepsTimer.start();
		else
			xepsTimer.invalidate();
	}

	if (reply->error())
	{
		sendPublicMessage(QString("%1 — и от этого вы ждёте фичи?").arg(nick));
		return;
	}
	QMap<QString, QStringList> features;
	foreach (const QString &feature, reply->item().features())
	{
		QPair<QString, QString> info = xeps.value(feature);
		if (info.first.isEmpty())
			info.first = feature;
		if (info.second.isEmpty())
		{
			info.first = QString();
			info.second = feature;
		}
		if (info.second == "N/A")
			info.second = QString();
		features[info.second] += info.first;
	}
	QString text = nick + ":\n";
	QMap<QString, QStringList>::iterator it;
	for (it = features.begin(); it != features.end(); ++it)
	{
		if (!it.key().isEmpty())
			text += it.key() + "\n";
		QStringList descriptions = it.value();
		qSort(descriptions);
		descriptions.removeDuplicates();
		foreach (const QString &desc, descriptions)
		{
			if (desc.trimmed().isEmpty() || desc.trimmed().startsWith("See "))
				continue;
			text += " * " + desc + "\n";
		}
	}
	sendPrivateMessage(from, text);
}

void Bot::sendPublicMessage(const QString &text)
{
	room->send(text);
	/*Jreen::Message mes(Jreen::Message::Groupchat, Jreen::JID("talks@conference.qutim.org"), text);
	client->send( mes );*/
}

void Bot::sendPublicMessage(const QString &to, const QString &text)
{
	sendPublicMessage(to + ": " + text);
}

void Bot::sendPrivateMessage(const QString &to, const QString &text)
{
	Jreen::Message mes(Jreen::Message::Chat, Jreen::JID(to), text);
	client->send(mes);
}

void Bot::sendGroupMessage(const QString &to, const QString &text)
{
	Jreen::Message mes(Jreen::Message::Groupchat, Jreen::JID(to), text);
	client->send(mes);
}

void Bot::addDelayedMessage(const QString &from, const QString &to, const QString &message)
{
	delayedMessages[to]->enqueue({from, QDateTime::currentDateTime(), message});
	Settings::mutableInstance().set("delayed_messages", delayedMessages);
}

void Bot::processMessage(const Jreen::Message &message)
{
	if (message.body().isEmpty())
		return;

    if (message.body().startsWith("/me "))
        insertLogRecord("action", message.from().resource(), message.body().mid(4),
			stringOfRole(users[message.from().resource()].role));
    else
        insertLogRecord("message", message.from().resource(), message.body(),
			stringOfRole(users[message.from().resource()].role));
	/* Don't handle own messages and topic too*/
	if (message.from() == fullConferenceName())
		return;
	if (isBot(message.from().resource()))
		return;
#define SP "[_.…–— \\$\\*-]*"
	static QRegExp checkPython(
			"([^а-яёa-zґїіє]|^)п" SP "[иuіїєeе]" SP "т" SP "[оo]" SP "н([^а-яёa-zґїіє]|$)"
			"|" "([^а-яёa-zґїіє]|^)п" SP "[еєаea]" SP "й" SP "т" SP "[оo]" SP "н([^а-яёa-zґїіє]|$)"
			"|" "[pр]" SP "y" SP "t" SP "h" SP "[oо]" SP "[nп]",
			Qt::CaseInsensitive,
			QRegExp::RegExp2
			);
	static QRegExp checkAnime(
			"([^а-яёa-zґїіє]|^)[aаoо]" SP "н" SP "[eєеиіїu]" SP "[мm]" SP "[eєеэ]" "([^а-яёa-zґїіє]|$)",
			Qt::CaseInsensitive
			);
	static QRegExp checkTranslation("^[a-z]{2}\\|[a-z]{2}\\s");
	static QRegExp callBot("^(аарон|протопоп|брат|святой отец|bot|sudo|):? ");
	Jreen::Message m = message;
	QString s = m.body();

#ifdef WITH_MORPHO
	QString orthoError = checkOrtho(s);
	if (orthoError.isEmpty())
		orthoError = checkOrtho(message.from().resource());
	qDebug() << "Ortho check result is:" << orthoError;
	if (!orthoError.isEmpty())
	{
		if(canKick(message.from().resource()))
		{
			qDebug() << "I can kick" << message.from().resource();
			kick(message.from().resource(), orthoError);
		}
		else
			sendPublicMessage(message.from().resource(), orthoError);
		return;
	}
#endif

	if (checkPython.indexIn(s) != -1)
		s.prepend("python ");
	else if (checkAnime.indexIn(s) != -1)
		s.prepend("anime ");
	else if (checkTranslation.indexIn(s) != -1)
		s.prepend("translate ");
	s.replace(callBot, "");
	qDebug() << s;
	m.setBody(s);
	handlers->process(m);
#undef SP
}

void Bot::processSubject(const QString &subject, const QString &source)
{
	insertLogRecord("topic", source, subject);
}

void Bot::sendDelayedMessages(const QString &nick)
{
	static auto sendDelayedMessage = [this](const QString &nick, const DelayedMessage &msg)
	{
		this->sendPublicMessage(nick, QString("сообщение от %1 (%2): %3").arg(msg.from).arg(msg.timeAdded.toString("dd.MM.yyyy HH:mm:ss")).arg(msg.message));
	};

	Jreen::JID j = room->realJid(nick);
	QString jid;
	if (j.isValid())
		jid = j.bare();
	else
	{
		jid = room->realJid(currentNicks[nick]).bare();
		if (jid.isNull())
		{
			jid = room->realJid(currentNicks[nicksHistory[nick]]).bare();
			if (jid.isNull())
				return;
		}
	}
	const QString &currentNick = currentNicks[jid];
	for (const DelayedMessage& msg : *delayedMessages[nick])
		sendDelayedMessage(currentNick, msg);
	delayedMessages[nick]->clear();
	delayedMessages.remove(nick);
	for (const DelayedMessage& msg : *delayedMessages[jid])
		sendDelayedMessage(currentNick, msg);
	delayedMessages[jid]->clear();
	delayedMessages.remove(jid);
	Settings::mutableInstance().set("delayed_messages", delayedMessages);
}

QHash<QString, int> Bot::pendingMessages() const
{
	QHash<QString, int> result;
	for (auto it = delayedMessages.begin(); it != delayedMessages.end(); ++it)
		result[it.key()] = it.value()->size();
	return result;
}

void Bot::forgetDelayedMessages(const QString &to)
{
	delayedMessages.remove(to);
	Settings::mutableInstance().set("delayed_messages", delayedMessages);
}

void Bot::userOut(const QString &type, const Jreen::Presence &presence, const QString &msg)
{
	insertLogRecord(type, presence.from().resource(), msg, stringOfRole(users[presence.from().resource()].role));
	users.remove(presence.from().resource());
}

void Bot::userIn(const Jreen::Presence &presence, const Jreen::MUCRoom::Participant *participant)
{
	QString message = QString::fromUtf8("%1 и %2 (%3)")
		.arg(stringOfAffiliation(participant->affiliation()))
		.arg(stringOfRole(participant->role()))
		.arg(statusOfPresence(presence));
	insertLogRecord("join", presence.from().resource(), message, stringOfRole(participant->role()));
	users[presence.from().resource()] = {
		participant->affiliation(),
		participant->role(),
		presence.subtype(),
		presence.status(),
	};
}

void Bot::processPresence(const Jreen::Presence &presence, const Jreen::MUCRoom::Participant *participant)
{
	if(participant) // user in conference
	{
		if(participant->isBanned())
			userOut("ban", presence, participant->reason());
		else if(participant->isKicked())
			userOut("kick", presence, participant->reason());
		else if(participant->isNickChanged())
		{
			QString oldNick = presence.from().resource();
			QString newNick = participant->newNick();
			insertLogRecord("nick", oldNick, newNick, stringOfRole(participant->role()));
			users[newNick] = users.take(oldNick);;
		}
		else if(participant->role() == Jreen::MUCRoom::RoleNone)
			userOut("left", presence, "");
		else if(!users.contains(presence.from().resource()))
			userIn(presence, participant);
		else
		{
			auto& u = users[presence.from().resource()];
			if(participant->affiliation() != u.aff)
				insertLogRecord("affiliation", presence.from().resource(),
						stringOfAffiliation(participant->affiliation()),
						stringOfRole(participant->role()));
			if(participant->role() != u.role)
				insertLogRecord("role", presence.from().resource(),
						stringOfRole(participant->role()),
						stringOfRole(participant->role()));
			if(presence.subtype() != u.status || presence.status() != u.message)
			{
				QString status = statusOfPresence(presence);
				if(!status.isNull())
					insertLogRecord("presence", presence.from().resource(),
							statusOfPresence(presence),
							stringOfRole(participant->role())
							);
			}
			u.aff = participant->affiliation();
			u.role = participant->role();
			u.status = presence.subtype();
			u.message = presence.status();
		}
	}

	if (!participant || !participant->isJoined())
		return;
	const QString &nick = presence.from().resource();
	const QString &jid = participant->realJID().bare();
	currentNicks[jid] = nick;
	nicksHistory[nick] = jid;
	Settings::mutableInstance().set("current_nicks", currentNicks);
	Settings::mutableInstance().set("nicks_history", nicksHistory);
	sendDelayedMessages(nick);
}
