/****************************************************************************
 * bot.h
 *  Copyright © 2011, Vsevolod Velichko.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#ifndef BOT_S4748RP2
#define BOT_S4748RP2

#include <QObject>
#include <QHash>
#include <QElapsedTimer>
#include <QQueue>
#include <QSqlDatabase>
#include <QSqlQuery>
#include <jreen/client.h>
#include <jreen/abstractroster.h>
#include <jreen/mucroom.h>
#include "options.h"
#include "delayedmessage.h"

class Handlers;

class Bot : public QObject
{
	Q_OBJECT
	private:
		typedef void (Bot::*Handler)(Jreen::Message&);

		Jreen::Client *client;
		Jreen::AbstractRoster *roster;
		Jreen::MUCRoom *room;
		Handlers *handlers;

		struct PingRequest
		{
			PingRequest(const QString &j) : nick(j) { timer.start(); }
			QString nick;
			QElapsedTimer timer;
		};
		struct RosterItem
		{
			Jreen::MUCRoom::Affiliation aff;
			Jreen::MUCRoom::Role role;
			Jreen::Presence::Type status;
			QString message;
		};

		QElapsedTimer xepsTimer;
		QHash<QString, QPair<QString, QString> > xeps;
		QHash<QString, PingRequest*> pings;
		QHash<QString, QString> simpleRequests;
		QHash<QString, DelayedMessages> delayedMessages;
		QHash<QString, QString> currentNicks; // jid -> nick
		QHash<QString, QString> nicksHistory; // nick -> jid
		QHash<QString, RosterItem> users;

		QSqlDatabase db;


		void userOut(const QString &type, const Jreen::Presence &presence, const QString &msg);
		void userIn(const Jreen::Presence &presence, const Jreen::MUCRoom::Participant *participant);
	public:
		enum OwnerStatus
		{
			NonOwner,
			Owner,
			SuperOwner
		};

		Bot();
		inline QString conferenceNick() const { return Options::instance()["conferenceNick"].toString(); }
		inline QString conferenceName() const { return Options::instance()["conferenceName"].toString(); }
		inline QString fullConferenceName() const { return conferenceName() + "/" + conferenceNick(); }
		inline bool canKick(const QString &nick) const { return room->canKick(nick); }
		inline bool canBan(const QString &nick) const { return room->canBan(nick); }
		QHash<QString, int> pendingMessages() const;
	public slots:
		OwnerStatus getOwnerStatus(const QString& nick);
		bool isBot(const QString& nick);
		void sendPublicMessage(const QString &text);
		void sendPublicMessage(const QString &to, const QString &text);
		void sendPrivateMessage(const QString &to, const QString &text);
		void sendGroupMessage(const QString &to, const QString &text);
		void onConnected();
		void onDisconnected(Jreen::Client::DisconnectReason reason);
		void joinGroupchat();
		void sendPing(const QString &nick);
		void sendVersionRequest(const QString &nick);
		void sendVcardRequest(const QString &nick);
		void sendTimeRequest(const QString &nick);
		void sendInfoRequest(const QString &nick, const Jreen::JID &from);
		void onPingResult(const Jreen::IQ& response, int);
		void onVersionResult(const Jreen::IQ& response, int);
		void onVcardResult(const Jreen::IQ& response, int);
		void onTimeResult(const Jreen::IQ& response, int);
		void onInfoResult();
		void processMessage(const Jreen::Message &message);
		void processPresence(const Jreen::Presence &presence, const Jreen::MUCRoom::Participant *participant);
		void processSubject(const QString &subject, const QString &author);
		bool isPresent(const QString &nick);
		void ban(const QString &nick);
		void unban(const QString &jid);
		void kick(const QString &nick, const QString &reason = "Quisque punietur propter fidem suam.");
		void setModer(const QString &nick);
		void addDelayedMessage(const QString &from, const QString &to, const QString &message);
		void sendDelayedMessages(const QString &nick);
		void forgetDelayedMessages(const QString &to);
	signals:
		void ready();
};

#endif /* end of include guard: BOT_S4748RP2 */
