/****************************************************************************
 * buildtracker.cpp
 *  Copyright © 2015, Nicolay Izoderov <nico-izo@ya.ru>.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#include "shortener.h"
#include "buildtracker.h"
#include "nocachingcache.h"
#include "options.h"
#include "settings.h"
#include <QNetworkRequest>
#include <QNetworkReply>
#include <QJsonDocument>
#include <QJsonArray>
#include <QJsonObject>
#include <QJsonValue>
#include <QDebug>

QString BuildInfo::toString(Shortener *shortener) const
{
	auto dt = QString(datetime);
	dt.truncate(19);
	return QString("[%1] %name%: %2 %3")
		.arg(dt)
		.arg(status)
		.arg(shortener->shorten(webLink));
}

BuildTracker::BuildTracker(Shortener *shortener, QObject *parent)
	: QObject(parent)
	, shortener(shortener)
{
	netman = new QNetworkAccessManager(this);
	netman->setCache(new NoCachingCache(this));

	connect(netman, SIGNAL(finished(QNetworkReply*)), this, SLOT(onReplyFinished(QNetworkReply*)));

	lastUpdated = Settings::instance().get<QHash<QString,QDateTime> >("last_updated");

	QMap<QString, QVariant> links = Options::instance()["buildLinks"].toMap();
	for (auto it = links.constBegin(); it != links.constEnd(); ++it) {
		repos[it.key()] = it.value().toString();
		if(!lastUpdated.contains(it.key()))
			lastUpdated[it.key()] = QDateTime::currentDateTimeUtc();
	}
	updateRepos();

	updater.setInterval(60000);
	connect(&updater, SIGNAL(timeout()), this, SLOT(updateRepos()));
	updater.start();
}

BuildTracker::~BuildTracker()
{
	netman->deleteLater();
}

void BuildTracker::updateRepos()
{
	for(auto it = repos.constBegin(); it != repos.constEnd(); ++it) {
		qDebug() << "Requesting build info " << it.key();
		auto reply = netman->get(QNetworkRequest(QUrl(it.value())));
		reply->setProperty("buildName", it.key());
	}
}

/**
 * \see https://launchpad.net/+apidoc/1.0.html#build
 *
 * @param jsonstr
 * @return List of BuildInfo structs
 */
QList<BuildInfo> BuildTracker::jsonToInfoList(const QByteArray& jsonstr)
{
	qDebug() << "Received JSON" << jsonstr;

	QJsonDocument doc = QJsonDocument::fromJson(jsonstr);
	Q_ASSERT(doc.isObject());
	Q_ASSERT(doc.object().value("entries").isArray());

	QJsonArray arr = doc.object()["entries"].toArray();

	QList<BuildInfo> ret;

	for(auto i = arr.constBegin(); i != arr.constEnd(); ++i) {
		BuildInfo b;
		QJsonObject object = (*i).toObject();

		b.datetime = object.value("datebuilt").toString();
		b.status = object.value("buildstate").toString();
		b.webLink = object.value("web_link").toString();
		ret.append(b);
	}

	return ret;
}

void BuildTracker::onReplyFinished(QNetworkReply *reply)
{
	QString name = reply->property("buildName").toString();
	QList<BuildInfo> infos = jsonToInfoList(reply->readAll());

	builds[name] = infos;

	qDebug() << "Adding infolist to builds[].";
	QDateTime maxT;
	maxT.setTimeSpec(Qt::UTC);
	maxT.setTime_t(0);

	for(int i = 0; i < infos.length(); ++i) {
		BuildInfo cur = infos.at(i);
		QString dt = cur.datetime;
		dt.truncate(19);
		QDateTime f = QDateTime::fromString(dt, "yyyy-MM-ddTHH:mm:ss");
		f.setTimeSpec(Qt::UTC);
		maxT = std::max(maxT, f);

		// If I emit signal too early, it causes segfault in jreen
		if(f > lastUpdated[name])
			emit newBuild(name, cur.toString(shortener));
	}

	lastUpdated[name] = std::max(maxT, QDateTime::currentDateTimeUtc());

	Settings::mutableInstance().set("last_updated", lastUpdated);

	reply->deleteLater();
}

// TODO max count and by default return only last
QString BuildTracker::getRepoMessage(const QString &repo)
{
	if (!repos.contains(repo))
		return QString();

	QString str;

	for(auto k = builds[repo].begin(); k != builds[repo].end(); ++k)
		str += k->toString(shortener).replace(QLatin1String("%name%"), repo) + "\n";

	return str;
}
