/****************************************************************************
 * gittracker.cpp
 *  Copyright © 2011, Vsevolod Velichko.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#include "gittracker.h"
#include "options.h"
#include "nocachingcache.h"
#include "shortener.h"
#include <QRegExp>
#include <QProcess>
#include <QProcessEnvironment>
#include <QDebug>
#include <QUrl>
#include <QXmlQuery>
#include <QDateTime>

GitTracker::GitTracker(Shortener *b, QObject *p) :
	QObject(p),
	shortener(b),
	nam(NULL)
{
	nam = new QNetworkAccessManager(this);
	nam->setCache(new NoCachingCache(this));
	QMap<QString, QVariant> folders = Options::instance()["gitRss"].toMap();
	QMap<QString, QVariant> links = Options::instance()["gitLinks"].toMap();
	for (auto it = folders.constBegin(); it != folders.constEnd(); ++it)
		if (links.contains(it.key()))
			repos.insert(it.key(), {QString(), it.value().toString(), links[it.key()].toString(), QString(), QString(), QString()});
	updateRepos();
	updater.setInterval(60000);
	connect(&updater, SIGNAL(timeout()), this, SLOT(updateRepos()));
	updater.start();
}

QString GitTracker::getRepoMessage (const QString &repo)
{
	qDebug() << repo;
	if (!repos.contains(repo))
		return QString();
	if (repos[repo].link.isEmpty())
	{
		repos[repo].link = shortener->shorten(repos[repo].longLink);
		repos[repo].commitMessage += "\nDiff: " + repos[repo].link;
	}
	qDebug() << repos[repo].rssLink;
	qDebug() << repos[repo].link;
	qDebug() << repos[repo].commitMessage;
	return repos[repo].commitMessage;
}

void GitTracker::updateRepos()
{
	for(auto it = repos.begin(); it != repos.end(); ++it)
	{
		qDebug() << "Looking in " << it.value().rssLink;
		QUrl request(it.value().rssLink);
		QXmlQuery query;
		query.setNetworkAccessManager(nam);
		QString results;
		query.setFocus(request);

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:updated/text()"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString updated = results.trimmed();
		if (updated < it.value().updated)
			continue;

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:entry[1]/a:id/text()"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString id = results.trimmed();

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:entry[1]/a:title/text()"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString title = results.trimmed().replace("&lt;", "<").replace("&gt;", ">");
		qDebug() << "Title:" << title;

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:entry[1]/a:link/string(@href)"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString link = results.trimmed();
		qDebug() << "Link:" << link;

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:entry[1]/a:author/a:name/text()"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString author = results.trimmed();
		qDebug() << author;

		query.setQuery(
				"declare namespace media='http://search.yahoo.com/mrss/';"
				"declare namespace a='http://www.w3.org/2005/Atom';"
				"a:feed/a:entry[1]/a:updated/text()"
			      );
		if (!query.evaluateTo(&results))
			continue;
		QString updateTime = QDateTime::fromString(results.trimmed(), Qt::ISODate).toString("dd.MM.yyyy hh:mm:ss");
		qDebug() << updateTime;

		QString m = QString("%1 on %2:\n%4").arg(author).arg(updateTime).arg(title);
		it.value().commitMessage = m;
		it.value().updated = updated;
		if (it.value().longLink != link)
		{
			it.value().longLink = link;
			it.value().link = QString();
		}
		if (!it.value().id.isEmpty() && (it.value().id != id))
		{
			emit newCommit(it.key(), getRepoMessage(it.key()));
		}
		it.value().id = id;
	}
}
