/****************************************************************************
 * handlers.cpp
 *  Copyright © 2011, Vsevolod Velichko.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#include "handlers.h"
#include "options.h"
#include "bot.h"
#include "shortener.h"
#include "gittracker.h"
#include "google.h"
#include "buildtracker.h"
#include <QDebug>
#include <QDate>
#include <QProcess>
#include <QCoreApplication>

static inline QString repositories()
{
	QMap<QString, QVariant> descriptions = Options::instance()["gitDescriptions"].toMap();
	QString repos = "доступные хранилища:\n";
	for (auto it = descriptions.constBegin(); it != descriptions.constEnd(); ++it)
		repos += QString("%1 - %2\n").arg(it.key()).arg(it.value().toString());
	return repos;
}

struct AbstractHandlerHelper
{
	virtual void operator()(Jreen::Message &msg) = 0;
};
template <typename Method> struct HandlerHelper : public AbstractHandlerHelper
{
	HandlerHelper(Method m) : method(m) {}
	void operator()(Jreen::Message &msg) { method(msg); }
	Method method;
};
class Handler
{
	public:
		template <typename Method> Handler(Method method)
		{
			helper = new HandlerHelper<Method>(method);
		}
		void operator()(Jreen::Message &msg) { (*helper)(msg); }
		AbstractHandlerHelper *helper;
};
class HandlersPrivate
{
	Q_DECLARE_PUBLIC(Handlers);
	private:
		Bitly *bitly;
		Clckru *clckru;
		Bot *bot;
		Google *google;
		BuildTracker *builds;
		GitTracker *repos;
	public:
		Handlers *q_ptr;

		HandlersPrivate(Bot *b)
			: bitly(new Bitly(Options::instance()["bitlyLogin"].toString(), Options::instance()["bitlyPassword"].toString()))
			, clckru(new Clckru)
			, bot(b)
			, google(new Google(Options::instance()["googleReferer"].toString()))
			, builds(nullptr)
			, repos(new GitTracker(clckru, NULL))
		{
		}
		virtual ~HandlersPrivate()
		{
			delete repos;
			delete bitly;
			delete google;
			delete builds;
			delete clckru;
		}
		void setQ(Handlers *q)
		{
			q_ptr = q;
			repos->setParent(q);
		}
		void echo(const QString &text)
		{
			bot->sendPublicMessage(text);
		}
		void say(const QString &text, Jreen::Message &message)
		{
			bot->sendPublicMessage(message.from().resource(), text);
		}
		void say(const QString &text, const QString &target)
		{
			bot->sendPublicMessage(target, text);
		}
		void getGit(const QString &repo, Jreen::Message &message)
		{
			say(repos->getRepoMessage(repo), message);
		}
		void getBuild(const QString &repo, Jreen::Message &message)
		{
			say(builds->getRepoMessage(repo), message);
		}
		void shorten(Jreen::Message &m)
		{
			say(clckru->shorten(m.body()), m);
		}
		void expand(Jreen::Message &m)
		{
			say(bitly->expand(m.body()), m);
		}
};

Handlers::Handlers(Bot *p = NULL) :
	QObject(p),
	d_ptr(new HandlersPrivate(p)),
	bot(p)
{
	Q_D(Handlers);
	d->setQ(this);
	connect(d->repos, SIGNAL(newCommit(QString,QString)), SLOT(onNewCommit(QString,QString)));

#define F(fun) new Handler([d](Jreen::Message &m){fun;})
#define SAY(text, always) F(if (m.body().isEmpty() && !always) return; d->say(text, m))
#define GIT(name) F(d->getGit(#name, m))
#define BUILD(name) F(d->getBuild(#name, m))
#define BITLY(fun) F(if (m.body().isEmpty()) return; d->say(d->bitly->fun(m.body()), m))
	handlers["!short"] = handlers["!скукожь"] = handlers["!shorten"] =
			F(
				d->shorten(m);
			);
	handlers["!enlarge"] = handlers["!раскукожь"] = handlers["!expand"] = BITLY(expand);
	handlers["аминь"] = SAY("воистину аминь", true);
	handlers["!логи"] = SAY(Options::instance()["logsRoot"].toString() + "today", true);
	handlers["!вчера"] = SAY(Options::instance()["logsRoot"].toString() + "yesterday", true);
	handlers["git"] = F(if (!m.body().isEmpty()) return; d->say(repositories(), m));
	/**
	 * @todo invent some way to create git retrieval handlers automatically, based on config
	 */
	handlers["tgit"] = GIT(tgit);
	handlers["jgit"] = GIT(jgit);
	handlers["egit"] = GIT(egit);
	handlers["ogit"] = GIT(ogit);
	handlers["sbzr"] = GIT(sbzr);
	handlers["sbuild"] = BUILD(sbuild);
	handlers["ubuild"] = BUILD(ubuild);
	handlers["testbuild"] = BUILD(testbuild);
	handlers["python"] = SAY("покайся, грешник! Никогда не произноси этого слова! Пейтонофилы будут гореть в аду!", true);
	handlers["anime"] = SAY("окстись, несчастный! Онемэ - удел содомитов и мужеложцев!", true);
	handlers["!опять"] = handlers["!снова"] = handlers["!again"] = SAY("https://dl.dropboxusercontent.com/u/27515337/qutim.humour.png", true);
	handlers["!stats"] = handlers["!статистика"] = SAY("https://metrika.yandex.ru/stat/?id=5541727", true);
	handlers["."] = handlers["!!"] =
		F(
			QString target = m.body();
			if (target == d->bot->conferenceNick())
			{
				d->say("я отвечаю на запросы мгновенно, с Божьей помощью, ведомый рукой Господа.", m);
				return;
			}
			if (target.isEmpty())
				target = m.from().resource();
			d->bot->sendPing(target);
		 );
	handlers["!v"] = handlers["!м"] =
		F(
			QString target = m.body();
			if (target == d->bot->conferenceNick())
			{
				d->say("аз есмь Альфа и Омега.", m);
				return;
			}
			if (target.isEmpty())
				target = m.from().resource();
			d->bot->sendVersionRequest(target);
		 );
	handlers["!rand"] = handlers["!рандом"] = handlers["!random"] =
		F(
			QStringList l = m.body().split(' ', QString::SkipEmptyParts);
			if (l.isEmpty())
				return;
			d->say(
				"советую выбрать: " + l.at(qrand() % l.size()),
				m);
		 );
	handlers["!дай"] =
		F(
			if(d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
				return;
			if(m.body().section(' ', 0, 0) == "модера")
				d->bot->setModer(m.from().resource());
		 );
	handlers["!сжечь"] = handlers["!kick"] =
		F(
			const QString &target = m.body();
			if ((target == d->bot->conferenceNick()) || (target == d->bot->fullConferenceName()))
			{
				d->say("глупцы сгорают первыми", m);
				d->bot->kick(m.from().resource());
				return;
			}
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
			{
				d->say("тебе — не положено, молись лучше усерднее.", m);
				return;
			}
			if (d->bot->getOwnerStatus(target) != Bot::NonOwner)
				d->say("может, лучше сжечь тебя?", m);
			else if (d->bot->canKick(target))
				d->bot->kick(target);
			else
				d->say("не послал мне Господь достаточно сил, а то бы обязательно.", m);
		 );
	handlers["!сварить"] = handlers["!колесовать"] = handlers["!ban"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
			{
				d->say("тебе — не положено, молись лучше усерднее.", m);
				return;
			}
			const QString &target = m.body();
			if ((target == d->bot->conferenceNick()) || (target == d->bot->fullConferenceName()))
			{
				d->say("глупцы сгорают первыми", m);
				d->bot->kick(m.from().resource());
				return;
			}
			if (d->bot->getOwnerStatus(target) != Bot::NonOwner)
				d->say("может, лучше колесовать тебя?", m);
			else if (d->bot->canBan(target))
				d->bot->ban(target);
			else
			{
				d->bot->ban(target);
				d->say("попробовал, а там уж на всё воля Божья.", m);
				//d->say("не послал мне Господь достаточно сил, а то бы обязательно.", m);
			}
		 );
	handlers["!разбанить"] = handlers["!unban"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
			{
				d->say("тебе — не положено, молись лучше усерднее.", m);
				return;
			}
			const QString &target = m.body();
			d->bot->unban(target);
			d->say("попробовал, а там уж на всё воля Божья.", m);
		 );
	handlers["скажи"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
				return;
			if (m.body().section(' ', 0, 0).indexOf('@') != -1)
				d->bot->sendGroupMessage(m.body().section(' ', 0, 0), m.body().section(' ', 1));
		 );
	handlers["!sh"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) != Bot::SuperOwner)
				return;
			QProcess proc;
			proc.start(m.body());
			proc.waitForFinished();
			d->say(QString::fromUtf8(proc.readAllStandardOutput().constData()).trimmed(), m);
		 );
	handlers["!литургия"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) != Bot::SuperOwner)
				d->say("всё врёшь, окаянный, не началась ещё!", m);
			else
			{
				d->say("спешу молиться!", m);
				QCoreApplication::processEvents();
				QTimer::singleShot(3000, QCoreApplication::instance(), SLOT(quit()));
			}
		 );
	handlers["!vcard"] = handlers[".vcard"] = handlers["!вкард"] =
		F(
			QString target = m.body();
			if (target == d->bot->conferenceNick())
			{
				d->say(Options::instance()["botInfoLink"].toString(), m);
				return;
			}
			if (target.isEmpty())
				target = m.from().resource();
			d->bot->sendVcardRequest(target);
		 );
	handlers["!translate"] =
		F(
			QString langpair = m.body().section(' ', 0, 0);
			QString text = m.body().section(' ', 1);
			if (text.isEmpty())
			{
				d->say("что переводить-то?", m);
				return;
			}
			QString translation = d->google->translate(langpair, text);
			if (translation.isEmpty())
				translation = "непереводимая игра слов, дитя моё.";
			d->say(translation, m);
		 );
	handlers["!фичи"] =
		F(
			QString target = m.body();
			if (target == d->bot->conferenceNick())
			{
				d->say("Сокрыты они от очей всего живущего и от птиц небесных утаены (Иов 28:21).", m);
				return;
			}
			if (target.isEmpty())
				target = m.from().resource();
			d->bot->sendInfoRequest(target, m.from());
		);
	handlers["!передать"] =
		F(
			QString target = m.body().startsWith('"') ? m.body().section('"', 0, 0, QString::SectionSkipEmpty) : m.body().section(' ', 0, 0);
			QString body = m.body().startsWith('"') ? m.body().section('"', 1, -1, QString::SectionSkipEmpty) : m.body().section(' ', 1);
			if(target.isEmpty() || body.isEmpty())
			{
				d->say("что передать-то?", m);
				return;
			}
			d->bot->addDelayedMessage(m.from().resource(), target, body);
			if(d->bot->isPresent(target))
			{
				d->bot->sendDelayedMessages(target);
				return;
			}
			d->say("при случае обязательно передам", m);
		 );
	handlers["!письма"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
				return;
			auto&& messages = d->bot->pendingMessages();
			if (!messages.size())
			{
				d->say("Писем нет.", m);
				return;
			}
			for (auto it = messages.begin(); it != messages.end(); ++it)
				d->say(QString::fromUtf8("%1: %2").arg(it.key()).arg(it.value()), m);
		 );
	handlers["!забудь"] =
		F(
			if (d->bot->getOwnerStatus(m.from().resource()) == Bot::NonOwner)
				return;
			QString target = m.body().startsWith('"') ? m.body().section('"', 0, 0, QString::SectionSkipEmpty) : m.body().section(' ', 0, 0);
			d->bot->forgetDelayedMessages(target);
		 );
	handlers["breaking_news.jpg"] = SAY("http://habrastorage.org/storage2/114/d36/d1f/114d36d1f4502e85b3580ae8f62c7cad.gif", true);
	handlers["!часики"] = handlers["!time"] = handlers["!date"] = handlers["!datetime"] = handlers["!часы"] = handlers["!время"] =
		F(
			QString target = m.body();
			if (target == d->bot->conferenceNick())
			{
				d->say("время разрушать, и время строить", m);
				return;
			}
			if (target.isEmpty())
				target = m.from().resource();
			d->bot->sendTimeRequest(target);
		 );
#undef BITLY
#undef GIT
#undef SAY
#undef F
}

void Handlers::process(Jreen::Message &message)
{
	QString command = message.body().section(' ', 0, 0);
	if (handlers.contains(command))
	{
		qDebug() << "Processing command" << command;
		message.setBody(message.body().section(' ', 1));
		(*handlers[command])(message);
	}
	else
	{
		qDebug() << "Unknown command" << command;
	}
}

void Handlers::parsePingResult(const QString &nick, qint64 msecs, bool isError)
{
	Q_D(Handlers);
	typedef QPair<QString, QString> Planet;
	static QMap<double, Planet> planets;
	if (planets.empty())
	{
		planets[2.6] = Planet("Луне", "Луной");
		planets[276] = Planet("Венере", "Венерой");
		planets[522] = Planet("Марсе", "Марсом");
		planets[610] = Planet("Меркурии", "Меркурием");
		planets[29334] = Planet("Плутоне", "Плутоном");
	}
	if (isError)
		d->echo(QString("дитя божье %1 не может дать ответа на пинг, а узнал я это за %2 мс").arg(nick).arg(msecs));
	else if(msecs < 2600) // 1.3с — расстояние до Луны
		d->echo(QString("дитя божье %1 среагировало за %2 мс").arg(nick).arg(msecs));
	else
	{
		auto upper = planets.upperBound(msecs/1000.);
		if (upper == planets.end())
		{
			d->echo(QString("дитя божье %1 не иначе как вознеслось прямо в рай (пинг %2 мс, %3 млн. км от Земли)").arg(nick).arg(msecs).arg(QString::number(msecs*0.0015, 'f', 3)));
			return;
		}
		auto lower = upper - 1;
		if (msecs/1000. - lower.key() < 2)
			d->echo(QString("дитя божье %1 находится где-то на %2 (пинг %3 мс)").arg(nick).arg(lower.value().first).arg(msecs));
		else if (upper.key() - msecs/1000. < 2)
			d->echo(QString("дитя божье %1 находится где-то на %2 (пинг %3 мс)").arg(nick).arg(upper.value().first).arg(msecs));
		else
			d->echo(QString("дитя божье %1 находится где-то между %2 и %3 (пинг %4 мс, %5 млн. км от Земли)").arg(nick).arg(lower.value().second).arg(upper.value().second).arg(msecs).arg(QString::number(msecs*0.00015, 'f', 3)));


	}
}
void Handlers::parseVersionResult(QString nick, QString name, QString version, QString os)
{
	Q_D(Handlers);
	if (name.isEmpty())
		name = "какую-то херню";
	if (os.isEmpty())
		os = "какой-то херни";
	d->echo(QString("дитя божье %1 использует %2 %3 под управлением %4").arg(nick).arg(name).arg(version).arg(os));
}
void Handlers::parseVcardResult(QString nick, Jreen::VCard::Ptr vcard)
{
	Q_D(Handlers);
	if (!vcard)
	{
		d->echo(nick + "? Кто это?");
		return;
	}
	QString message("Прозвище в конференции: " + nick);
	if (!vcard->formattedName().isEmpty())
		message += "\nИмя: " + vcard->formattedName();
	if (!vcard->nickname().isEmpty())
		message += "\nЗемное прозвище: " + vcard->nickname();
	if (vcard->url().isValid())
		message += "\nАдрес земной: " + vcard->url().toString();
	if (vcard->birthday().isValid())
		message += "\nДень рождения: " + vcard->birthday().toString("dd-MM-yyyy");
	d->echo(message);
}

void Handlers::onConnected()
{
	Q_D(Handlers);
	d->builds = new BuildTracker(d->clckru);
	connect(d->builds, SIGNAL(newBuild(QString,QString)), SLOT(onNewBuild(QString,QString)));
}

void Handlers::onNewCommit(const QString &repo, const QString &message)
{
    Q_D(Handlers);
    d->echo(QString::fromUtf8("Новый коммит в %1:\n%2").arg(repo).arg(message));
}

void Handlers::onNewBuild(const QString &repo, const QString &message)
{
	Q_D(Handlers);
	QString cp = message;
	cp.replace(QLatin1String("%name%"), repo);
	d->echo(QString::fromUtf8("Новая сборка:\n %1").arg(cp));
}
