/****************************************************************************
 * logserver.cpp
 *  Copyright © 2014, Vsevolod Velichko.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#include "logserver.h"
#include "options.h"

#include <ctpp2/CTPP2StringOutputCollector.hpp>
#include <ctpp2/CTPP2SyscallFactory.hpp>
#include <ctpp2/CTPP2VMSTDLib.hpp>
#include <ctpp2/CTPP2VMFileLoader.hpp>
#include <ctpp2/CTPP2VM.hpp>
#include <ctpp2/CTPP2FileLogger.hpp>

#include <httpserver.hpp>

#include <iostream>
#include <memory>
#include <QVariant>
#include <QSqlQuery>
#include <QSqlError>
#include <QRegularExpression>
#include <QDate>
#include <QDebug>
#include <QCoreApplication>

#include <execinfo.h>
#include <signal.h>

#ifndef AARON_SHARE_PATH
#define AARON_SHARE_PATH ""
#endif

class UserFormatter : public CTPP::SyscallHandler
{
	public:
		UserFormatter() = default;
		virtual ~UserFormatter() = default;
		virtual INT_32 Handler(
				CTPP::CDT *argv,
				const UINT_32 argc,
				CTPP::CDT &retval,
				CTPP::Logger &logger
				)
		{
			if(argc != 2)
			{
				logger.Emerg("Usage: USER(role, name)");
				return -1;
			}

			const auto &role = argv[1].GetString();
			const auto &nick = argv[0].GetString();
			STLW::string result;
			result.reserve(55 + role.size() + nick.size());

			result.append("<span class=\"src\">");
			if(role == "член")
				result.append("+");
			else if(role == "модератор")
				result.append("@");
			result.append("<span class=\"nick_");
			result.push_back(nickHash(nick));
			result.append("\">");
			result.append(nick);
			result.append("</span></span>");

			retval = result;

			return 0;
		}

		virtual CCHAR_P GetName() const
		{
			return "user";
		}

	private:
		inline char nickHash(const std::string &str) const
		{
			// colors amount is intentionally limited to the
			// small number, which would fit into a subset
			// of printable ascii-chars and is prime
			static const char chars[] = "0123456789ab";
			uint32_t h = 0;
			for(const auto &c : str)
				h = 37 * h + (unsigned char)c;
			return chars[h % sizeof(chars)];
		}
};

struct CtppContext
{
	CtppContext()
	{
		factory.reset(new CTPP::SyscallFactory(100000));
		userUdf.reset(new UserFormatter);
		CTPP::STDLibInitializer::InitLibrary(*factory);
		factory->RegisterHandler(userUdf.get());

		loader.reset(new CTPP::VMFileLoader(AARON_SHARE_PATH "index.ct2"));
		core = loader->GetCore();
		basePath = Options::instance()["baseUrl"].toString().toStdString();
		yandexRca = Options::instance()["yandexRca"].toString().toStdString();
		yandexRcaRoot = Options::instance()["yandexRcaRoot"].toString().toStdString();
	}

	std::string basePath;
	std::string yandexRca;
	std::string yandexRcaRoot;
	std::unique_ptr<CTPP::SyscallFactory> factory;
	std::unique_ptr<CTPP::SyscallHandler> userUdf;
	std::unique_ptr<CTPP::VMFileLoader> loader;
	const CTPP::VMMemoryCore *core;
};

struct LogHandler : public httpserver::http_resource<LogHandler>
{
	std::shared_ptr<CtppContext> context;
	
	bool fillYearList(CTPP::CDT &dataHash)
	{
		QSqlQuery query("SELECT DISTINCT year FROM log");
		auto error = query.lastError();
		if (error.isValid())
		{
			std::cerr << "Error type: " << error.type() << std::endl;
			std::cerr << "Database: " << error.databaseText().toStdString() << std::endl;
			std::cerr << "Driver: " << error.driverText().toStdString() << std::endl;
			return false;
		}

		if (query.size() == 0)
			return false;

		auto &years = dataHash["years"] = CTPP::CDT(CTPP::CDT::ARRAY_VAL);

		while (query.next())
			years.PushBack(query.value(0).toInt());

		return true;
	}

	bool fillMonthList(CTPP::CDT &dataHash, int year)
	{
		QSqlQuery query;
		query.prepare("SELECT DISTINCT month FROM log WHERE year = :year");
		query.bindValue(":year", year);
		query.exec();

		auto error = query.lastError();
		if (error.isValid())
		{
			std::cerr << "Error type: " << error.type() << std::endl;
			std::cerr << "Database: " << error.databaseText().toStdString() << std::endl;
			std::cerr << "Driver: " << error.driverText().toStdString() << std::endl;
			return false;
		}

		if (query.size() == 0)
			return false;

		dataHash["year"] = year;
		auto &months = dataHash["months"] = CTPP::CDT(CTPP::CDT::ARRAY_VAL);
		while (query.next())
			months.PushBack(query.value(0).toInt());

		return true;
	}

	bool fillDayList(CTPP::CDT &dataHash, int year, int month)
	{
		QSqlQuery query;
		query.prepare("SELECT DISTINCT day FROM log WHERE year = :year AND month = :month");
		query.bindValue(":year", year);
		query.bindValue(":month", month);
		query.exec();

		auto error = query.lastError();
		if (error.isValid())
		{
			std::cerr << "Error type: " << error.type() << std::endl;
			std::cerr << "Database: " << error.databaseText().toStdString() << std::endl;
			std::cerr << "Driver: " << error.driverText().toStdString() << std::endl;
			return false;
		}

		if (query.size() == 0)
			return false;

		dataHash["year"] = year;
		dataHash["month"] = month;
		auto &days = dataHash["days"] = CTPP::CDT(CTPP::CDT::ARRAY_VAL);
		while (query.next())
			days.PushBack(query.value(0).toInt());

		return true;
	}

	bool fillMessages(CTPP::CDT &dataHash, int year, int month, int day)
	{
		QSqlQuery query;
		query.prepare("SELECT time, type, source, role, message FROM log WHERE year = :year AND month = :month AND day = :day ORDER BY time ASC");
		query.bindValue(":year", year);
		query.bindValue(":month", month);
		query.bindValue(":day", day);
		query.exec();

		auto error = query.lastError();
		if (error.isValid())
		{
			std::cerr << "Error type: " << error.type() << std::endl;
			std::cerr << "Database: " << error.databaseText().toStdString() << std::endl;
			std::cerr << "Driver: " << error.driverText().toStdString() << std::endl;
			return false;
		}

		if (query.size() == 0)
			return false;

		dataHash["year"] = year;
		dataHash["month"] = month;
		dataHash["day"] = day;

		QDate yesterday = QDate(year, month, day).addDays(-1);
		dataHash["y_year"] = yesterday.year();
		dataHash["y_month"] = yesterday.month();
		dataHash["y_day"] = yesterday.day();

		QDate tomorrow = QDate(year, month, day).addDays(1);
		if (tomorrow <= QDate::currentDate())
		{
			dataHash["t_year"] = tomorrow.year();
			dataHash["t_month"] = tomorrow.month();
			dataHash["t_day"] = tomorrow.day();
		}

		auto &messages = dataHash["messages"] = CTPP::CDT(CTPP::CDT::ARRAY_VAL);
		while (query.next())
		{
			auto msg = CTPP::CDT(CTPP::CDT::HASH_VAL);
			msg["time"] = query.value(0).toString().toStdString();
			msg["type"] = query.value(1).toString().toStdString();
			msg["source"] = query.value(2).toString().toHtmlEscaped().toStdString();
			msg["role"] = query.value(3).toString().toStdString();
			msg["message"] = wrapUrls(query.value(4).toString().toHtmlEscaped()).toStdString();
			messages.PushBack(msg);
		}

		return true;
	}

	QString wrapUrls(QString str)
	{
		static QRegularExpression urlRe(
                R"#(()#" // START PATTERN MATCH
				R"#((?:https?|s?ftp|xmpp|skype|mailto):(?://)?)#" // scheme
				R"#((?:[\w+-]+(?::[\w+-]+)?@)?)#" // user@password
				R"#([\w-]+(?:\.[\w-]+)*)#" // domain
				R"#(/[^\s"]*)#" // path+query+hash
                R"#())#" // END PATTERN MATCH
				,
				QRegularExpression::CaseInsensitiveOption
					| QRegularExpression::UseUnicodePropertiesOption
				);
		return str.replace(urlRe, R"#(<a class="ext" href="\1">\1</a>)#");
	}

	void render(const httpserver::http_request &req, httpserver::http_response **response)
	{
		std::string data;
		CTPP::StringOutputCollector collector(data);
		CTPP::FileLogger logger(stderr);

		UINT_32 instruction = 0;
		CTPP::CDT dataHash(CTPP::CDT::HASH_VAL);
		dataHash["basePath"] = context->basePath;
		dataHash["yandexRca"] = context->yandexRca;
		dataHash["yandexRcaRoot"] = context->yandexRcaRoot;

		CTPP::VM vm(context->factory.get(), 1 << 20, 1 << 20, 1 << 20);
		vm.Init(context->core, &collector, &logger);

		int year = 0, month = 0, day = 0;
		if (req.get_arg("year").size() > 0)
			year = atoi(req.get_arg("year").c_str());
		if (req.get_arg("month").size() > 0)
			month = atoi(req.get_arg("month").c_str());
		if (req.get_arg("day").size() > 0)
			day = atoi(req.get_arg("day").c_str());

		if (!year)
		{
			if (req.get_arg("today").size() > 0)
			{
				QDate ts = QDate::currentDate();
				year = ts.year();
				month = ts.month();
				day = ts.day();
			}
			if (req.get_arg("yesterday").size() > 0)
			{
				QDate ts = QDate::currentDate().addDays(-1);
				year = ts.year();
				month = ts.month();
				day = ts.day();
			}
		}
		std::cerr << "Year: " << year << " Month: " << month << " Day: " << day << std::endl;
		if (!year)
			fillYearList(dataHash);
		else if (!month)
			fillMonthList(dataHash, year);
		else if (!day)
			fillDayList(dataHash, year, month);
		else
			fillMessages(dataHash, year, month, day);

		vm.Run(context->core, &collector, instruction, dataHash, &logger);

		*response = new httpserver::http_response(httpserver::http_response_builder(data, 200, "text/html; charset=utf-8").string_response());
	}

};

void (*old_handler)(int);

void stop_everything(int signal)
{
	old_handler(signal);
	QCoreApplication::quit();
}

void log_access(const std::string &s)
{
	std::cerr << "ACCESS " << s << std::endl;
}
void log_error(const std::string &s)
{
	std::cerr << "ERROR " << s << std::endl;
}

void runLogServer()
{
	old_handler = signal(SIGINT, stop_everything);
	signal(SIGTERM, stop_everything);

	try
	{
		int port = Options::instance()["logsPort"].toInt();
		httpserver::webserver ws = httpserver::create_webserver(port)
		                           .max_threads(2)
		                           .debug()
		                           .log_access(&log_access)
		                           .log_error(&log_error)
		                           .no_basic_auth()
		                           .no_digest_auth()
		                           .no_ban_system()
		;
		LogHandler handler;
		handler.context.reset(new CtppContext);
		
		ws.register_resource("/", &handler, true);
		ws.register_resource("/{year|[0-9]+}", &handler, true);
		ws.register_resource("/{year|[0-9]+}/{month|[0-9]+}", &handler, true);
		ws.register_resource("/{year|[0-9]+}/{month|[0-9]+}/{day|[0-9]+}", &handler, true);
		ws.register_resource("/{year|[0-9]+}/{month|[0-9]+}/{day|[0-9]+}", &handler, true);
		ws.register_resource("/{today|today}", &handler, true);
		ws.register_resource("/{yesterday|yesterday}", &handler, true);
		ws.start(true);
	}
	catch (std::exception &e)
	{
		void *array[10];
		size_t size;
		size = backtrace(array, 10);
		backtrace_symbols_fd(array, size, STDERR_FILENO);
		throw;
	}
}
