/****************************************************************************
 * options.cpp
 *  Copyright © 2010, Vsevolod Velichko <torkvema@gmail.com>.
 *  Licence: GPLv3 or later
 *
 ****************************************************************************
 *                                                                          *
 *   This library is free software; you can redistribute it and/or modify   *
 *   it under the terms of the GNU General Public License as published by   *
 *   the Free Software Foundation; either version 3 of the License, or      *
 *   (at your option) any later version.                                    *
 *                                                                          *
 ****************************************************************************/

#include "options.h"
#include "3rdparty/qxt/qxtcommandoptions.h"
#include <boost/serialization/singleton.hpp>
#include <stdlib.h>
#include <QDebug>
#include <QFile>
#include <QDir>
#include <QCoreApplication>

Options::Options () :
	QObject(QCoreApplication::instance())
{
	// Default options
	options["botjid"] = "bot@xmpp.org/bot";
	options["botpassword"] = "password";
	options["conferenceNick"] = "nickname";
	options["conferenceName"] = "talks@conference.qutim.org";
}

Options::~Options ()
{
}

void Options::readCommandLine ()
{
	QxtCommandOptions parser;
	parser.add("jid", "Full JID of the bot", QxtCommandOptions::ValueRequired);
	parser.alias("jid", "j");
	parser.add("password", "Jabber password of the bot", QxtCommandOptions::ValueRequired);
	parser.alias("password", "p");
	parser.add("nick", "Nickname of the bot", QxtCommandOptions::ValueRequired);
	parser.alias("nick", "n");
	parser.add("conference", "Conference address", QxtCommandOptions::ValueRequired);
	parser.alias("conference", "c");
	parser.add("database", "SQLite Database filename", QxtCommandOptions::ValueRequired);
	parser.alias("database", "db");

	parser.add("help", "This help message");
	parser.alias("help", "?");
	parser.alias("help", "h");

	parser.parse(QCoreApplication::arguments());

	if (parser.count("help") || parser.showUnrecognizedWarning())
	{
		parser.showUsage();
		options["shouldExit"] = true;
	}
	if (parser.count("jid"))
		options["botjid"] = parser.value("jid");
	if (parser.count("password"))
		options["botpassword"] = parser.value("password");
	if (parser.count("nick"))
		options["conferenceNick"] = parser.value("nick");
	if (parser.count("conference"))
		options["conferenceName"] = parser.value("conference");
	if (parser.count("database"))
		options["database"] = parser.value("database");
}

void Options::parseConfig(const QString &path)
{
	static QRegExp commentStripper("###.*");
	static QRegExp whitespaceStripper1("^\\s+|\\s+$");
	static QRegExp whitespaceStripper2("\\s{2,}");
	static QRegExp parameterTuple("^([a-zA-Z_]+)\\s*=\\s*\"(.+)\"$");
	static QRegExp listElement("^([a-zA-Z_]+)\\[\\]\\s*=\\s*\"(.+)\"$");
	static QRegExp mapElement("^([a-zA-Z_]+)\\[([a-zA-Z_]+)\\]\\s*=\\s*\"(.+)\"$");

	if (!QFile::exists(path))
		return;
	QFile configFile(path);
	if (!configFile.open(QIODevice::ReadOnly|QIODevice::Text))
		return;
	QTextStream in(&configFile);
	while (!in.atEnd())
	{
		QString line = in.readLine();
		if (line
			.replace(commentStripper, "")
			.replace(whitespaceStripper1, "")
			.replace(whitespaceStripper2, " ")
			.isEmpty()
		   )
			continue;
		if (mapElement.indexIn(line) != -1)
		{
			QString name = mapElement.cap(1);
			qDebug() << "Loading map element" << mapElement.cap(2) << "into" << name << "with value:" << mapElement.cap(3);
			QMap<QString, QVariant> map = (isset(name) ? options[name].toMap() : QMap<QString, QVariant>());
			map[mapElement.cap(2)] = mapElement.cap(3);
			options[name] = map;
		}
		else if (listElement.indexIn(line) != -1)
		{
			QString name = listElement.cap(1);
			qDebug() << "Loading list element" << listElement.cap(2) << "into" << name;
			options[name] = (isset(name) ? options[name].toStringList() : QStringList()) << listElement.cap(2);
		}
		else if (parameterTuple.indexIn(line) != -1)
			options[parameterTuple.cap(1)] = parameterTuple.cap(2);
	}
}

void Options::readConfig()
{
	parseConfig(QDir::currentPath() + "/bot.conf");
	readCommandLine();
}

QVariant Options::operator[](const QString &name) const
{
	return options.value(name);
}

bool Options::isset(const QString &name) const
{
	return options.contains(name);
}

const Options& Options::instance()
{
	return boost::serialization::singleton<Options>::get_const_instance();
}

Options& Options::mutableInstance()
{
	return boost::serialization::singleton<Options>::get_mutable_instance();
}
