/****************************************************************************
 * ortho.cpp
 *  Copyright © 2011, Vsevolod Velichko.
 *  Licence: WTFPL with notes
 *
 ****************************************************************************
 *                                                                          *
 *              DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE                 *
 *                      Version 2, December 2004                            *
 *                                                                          *
 *   Copyright (C) 2004 Sam Hocevar <sam@hocevar.net>                       *
 *                                                                          *
 *   Everyone is permitted to copy and distribute verbatim or modified      *
 *   copies of this license document, and changing it as allowed as long    *
 *   as the name is changed.                                                *
 *                                                                          *
 *              DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE                 *
 *     TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION      *
 *                                                                          *
 *   0. You just DO WHAT THE FUCK YOU WANT TO.                              *
 *                                                                          *
 *   Notes:                                                                 *
 *   This piece of code is linked with the proprietary library.             *
 *   If you don't have that library, if you are GPL-minded man              *
 *   or just don't need these features,                                     *
 *   PLEASE, JUST DON'T FUCKING COMPILE THIS CODE                           *
 *                                                                          *
 ****************************************************************************/

#include <nigma_morph.h>
#include <QRegExp>
#include <QString>
#include <QByteArray>
#include <QStringList>
#include <QDebug>
#include <functional>

static morphology morph;

typedef std::function<QString(const QString &)> Checker;

auto generateSimpleChecker(const QString &regexp, const QString &replacement, const QString &requiredPartOfSpeech, const QString &message, const QString &deniedFormRegex = QString()) -> Checker
{
	return [=](const QString &word) -> QString
	{
		QRegExp search(regexp, Qt::CaseInsensitive);
		if (search.indexIn(word) == -1)
			return QString();
		QString wordFixed = word;
		wordFixed.replace(search, replacement);
		QByteArray fixed = wordFixed.toUtf8();
		if (morph.get_freq(fixed.constData(), true) == 0)
			return QString();
		char *partOfSpeech = morph.get_part_of_speech(fixed.constData(), true);
		if (!partOfSpeech)
			return QString();
		QByteArray _required = requiredPartOfSpeech.toUtf8();
		if (strncmp(partOfSpeech, _required.constData(), _required.size()) == 0)
		{
			free(partOfSpeech);
			if (!deniedFormRegex.isEmpty())
			{
				char *formNames = morph.get_form_names(fixed.constData(), true);
				bool banned = QRegExp(deniedFormRegex, Qt::CaseInsensitive).indexIn(QString::fromUtf8(formNames)) == -1;
				free(formNames);
				if (banned)
					return message.arg(word);
				else
					return QString();
			}
			return message.arg(word);
		}
		free(partOfSpeech);
		return QString();
	};
}

auto generateWordChecker(const QString &bannedWord) -> Checker
{
	return [=](const QString &word) -> QString
	{
		if (bannedWord.compare(word, Qt::CaseInsensitive) == 0)
			return QString::fromUtf8("> %1\nНет такого слова в русском языке!").arg(word);
		return QString();
	};
}

auto generateTextChecker(const QString &bannedText) -> Checker
{
	QRegExp re(bannedText, Qt::CaseInsensitive);
	return [=](const QString &text) -> QString
	{
		if (re.indexIn(text) != -1)
			return QString::fromUtf8("> %1\nЭто что вообще?").arg(re.cap());
		return QString();
	};
}

QString checkOrtho(const QString &message)
{
	static auto wordChecks = QList<Checker>()
		<< generateSimpleChecker(QString::fromUtf8("ТЬСЯ$"),
				QString::fromUtf8("ТСЯ"),
				QString::fromUtf8("Глагол"),
				QString::fromUtf8("> %1\nhttp://tsya.ru")
				)
		<< generateSimpleChecker(QString::fromUtf8("ТСЯ$"),
				QString::fromUtf8("ТЬСЯ"),
				QString::fromUtf8("Глагол"),
				QString::fromUtf8("> %1\nhttp://tsya.ru")
				)
		<< generateSimpleChecker(QString::fromUtf8("^НЕ"),
				QString::fromUtf8(""),
				QString::fromUtf8("Глагол"),
				QString::fromUtf8("> %1\n«Не» с глаголами пишется отдельно!\nhttp://gramota.ru/spravka/rules/?rub=def&text=19_24"),
				QString::fromUtf8("прич\\.")
				)
		<< generateWordChecker("вообщем")
		<< generateWordChecker("вообщемто")
		<< generateWordChecker("вобщем")
		<< generateWordChecker("будующий")
		<< generateWordChecker("ктото")
		<< generateWordChecker("чтото")
		<< generateWordChecker("както")
		<< generateWordChecker("гдето")
		<< generateWordChecker("когдато")
		<< generateWordChecker("кудато")
		<< generateWordChecker("ктонить")
		<< generateWordChecker("чтонить")
		<< generateWordChecker("гденить")
		<< generateWordChecker("когданить")
		<< generateWordChecker("куданить")
		<< generateWordChecker("андройд")
		<< generateWordChecker("изза")
		<< generateWordChecker("этож")
		<< generateWordChecker("влазиет")
		<< generateWordChecker("комуто")
		<< generateWordChecker("всетаки")
		<< generateWordChecker("догодались")
		<< generateWordChecker("пробывать")
		<< generateWordChecker("пробывал")
		<< generateWordChecker("попробывать")
		<< generateWordChecker("попробывал")
		;
	static auto textChecks = QList<Checker>()
		<< generateTextChecker("\\bс начало\\b")
		;
	qDebug() << "Checking ortho:" << message;
	char *m = morph.get_clean_text(message.toUtf8().constData(), true);
	QString text = QString::fromUtf8(m);
	QStringList words = text.split(' ', QString::SkipEmptyParts);
	free(m);
	qDebug() << "Words:" << words;
	QString response;
	for (auto &check : textChecks)
	{
		response = check(text);
		if (!response.isEmpty())
			return response;
	}
	for(QString &word : words)
	{
		QByteArray original = word.toUtf8();
		if (morph.get_freq(original.constData(), true))
			continue;
		for (auto &check : wordChecks)
		{
			response = check(word);
			if (!response.isEmpty())
				return response;
		}
	}
	return QString();
}
